#!/usr/bin/env python3

from pathlib import Path
from datetime import datetime
from pony.orm import Database, Required, Optional, Json, composite_key


db = Database()


@db.on_connect(provider="sqlite")
def _home_sqliterc(_, conn):
    rc = Path.home() / ".sqliterc"
    rc.exists() and conn.executescript(rc.read_text())


class Fsm(db.Entity):
    state = Required(str, 80)
    ts = Required(datetime, default=datetime.now)
    data = Optional(Json)
    composite_key(ts, state)


if __name__ == '__main__':
    db.bind('sqlite', filename=':memory:')
    db.generate_mapping(create_tables=True)



#!/usr/bin/env python3

from datetime import datetime, timedelta
from pony import orm
from .entities import Fsm, db

try:
    from loguru import logger
    log = logger.info
except ImportError:
    log = print

prefix_locked = ".locked-"


class NotFound(Warning): pass
class NotAllowed(Warning): pass


def new(state, data={}):
    with orm.db_session:
        i = Fsm(state=state, data=data)
    log(f"new {state} {i.id}")


@orm.db_session
def lock(state):
    if state.startswith(prefix_locked):
        raise NotAllowed(f"'{prefix_locked}*' is not allowed")
    ts = datetime.now() - timedelta(seconds=300)
    i = orm.select(
        i for i in Fsm if i.ts > ts and i.state == state
    ).order_by(Fsm.ts).for_update(skip_locked=True).first()
    if not i:
        raise NotFound(state)
    prev_info = i.to_dict()
    i.state = f"{prefix_locked}{i.state}"
    i.ts = datetime.now()
    delta = (i.ts - prev_info["ts"]).total_seconds()
    log(f"lock {state} {i.id} {delta}")
    return prev_info


@orm.db_session
def transit(id, state, data_patch=None):
    i = Fsm.get_for_update(id=id)
    if not i:
        raise NotFound(id)
    if not i.state.startswith(prefix_locked):
        raise NotAllowed(i.state, state)
    prev_state, prev_ts = i.state, i.ts
    i.state = state
    i.ts = datetime.now()
    if data_patch:
        i.data.update(data_patch)
    delta = (i.ts - prev_ts).total_seconds()
    log(f"transit {state} {i.id} {delta} {prev_state}")


@orm.db_session
def info(id):
    try:
        return Fsm[id].to_dict()
    except orm.ObjectNotFound:
        raise NotFound(id)


def _init_this():
    #orm.sql_debug(True)
    from os.path import abspath
    from yaml import safe_load
    try:
        options = safe_load(open("database.yaml"))
    except FileNotFoundError:
        options = {"provider": "sqlite", "filename": ":memory:"}
    fn = options.get("filename")
    if fn and fn != ":memory:":
        options["filename"] = abspath(fn)  # $CWD/filename
    db.bind(**options)
    db.generate_mapping(create_tables=True)


if __name__ == '__main__':
    db.bind('sqlite', filename=':memory:')
    db.generate_mapping(create_tables=True)
else:
    _init_this()



#!/usr/bin/env python3

from setuptools import setup

author = "lwzm"
name = "fsm-hub"

setup(
    name=name,
    version="1.2",
    description="Finite State Machine storage hub",
    author=author,
    author_email="{}@qq.com".format(author),
    url="https://github.com/{}/{}".format(author, name),
    keywords="fsm http web store".split(),
    packages=["fsm_hub"],
    scripts=[name],
    install_requires="pony falcon".split(),
    classifiers=[
        "Environment :: Console",
        "Programming Language :: Python :: 3",
    ],
)



#!/usr/bin/env python3

import json

from falcon import Request, Response, API, HTTPNotFound, HTTPForbidden, HTTPBadRequest, HTTP_CREATED

from .core import new, lock, transit, info, NotFound, NotAllowed


class InitNew:
    def on_post(self, req: Request, resp: Response, state):
        data = json.load(req.stream) if req.content_length else {}
        new(state, data)
        resp.status = HTTP_CREATED


class LockOne:
    def on_post(self, req: Request, resp: Response, state):
        try:
            item = lock(state)
        except NotAllowed as e:
            raise HTTPBadRequest(description=e.args[0])
        except NotFound:
            raise HTTPNotFound
        resp.body = json.dumps(item, default=str, ensure_ascii=False)


class TransitLocked:
    def on_post(self, req: Request, resp: Response, id, state):
        data = json.load(req.stream) if req.content_length else None
        try:
            transit(id, state, data)
        except NotFound:
            raise HTTPNotFound
        except NotAllowed as e:
            raise HTTPForbidden(description=" -> ".join(e.args))


class Info:
    def on_get(self, req: Request, resp: Response, id):
        try:
            item = info(id)
        except NotFound:
            raise HTTPNotFound
        resp.body = json.dumps(item, default=str, ensure_ascii=False)


application = API()
application.add_route('/new/{state}', InitNew())
application.add_route('/lock/{state}', LockOne())
application.add_route('/transit/{id:int}/{state}', TransitLocked())
application.add_route('/{id:int}', Info())


def main():
    from os import environ
    from wsgiref.simple_server import make_server
    make_server(
        environ.get("HOST", ""),
        int(environ.get("PORT", 1024)),
        application,
    ).serve_forever()


if __name__ == '__main__':
    main()
